//-----------------------------------------------------------------------------
// Copyright  2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		map_array_lib.h
//
// purpose	Define functions for the key and data type specified by
//		the source file that includes this file.
//-----------------------------------------------------------------------------

#include "map_lib.h"

//-----------------------------------------------------------------------------
// The following must be defined by the file that includes this one:
//
// map_type		Typedef this to the C type for the array type to be
//			used for the key or data, e.g. long double.
//
// MAP_TYPE_CODE	Define this to the int code value that identifies
//			this data type in a mixed data struct.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// struct	map_link
// struct	map_node
//
// purpose	Define structures needed to keep the AVL trees used to hold
//		the mapping.
//-----------------------------------------------------------------------------
struct map_link {
    avl_link		link	;
    size_t		keylen	;
    map_type		key	[1]; // array is dynamically extended
};
struct map_node {
    struct map_mixed	data	;
    struct map_link	link	;
};

//-----------------------------------------------------------------------------
// function	map_xx_length (internal, inline)
//
// purpose	Determine the length of a string of this key or data type.
//
// arguments	1 (const map_type *) string pointer
//
// returns	(size_t) length of string
//-----------------------------------------------------------------------------
inline static
size_t
map_xx_length (
    const map_type *	arg_str
) {
    size_t	len	;
    len = 0;
    while ( * arg_str ++ != (map_type) 0 ) ++ len;
    return len;
}

//-----------------------------------------------------------------------------
// function	map_xx_copy (internal, inline)
//
// purpose	Copy a string of this key or data type, and terminate it.
//
// arguments	1 (map_type *) destination pointer
//		2 (const map_type *) source pointer
//		3 (size_t) length to copy
//
// returns	(void) nothing
//-----------------------------------------------------------------------------
inline static
void
map_xx_copy (
    map_type *		arg_dst
    ,
    const map_type *	arg_src
    ,
    size_t		arg_len
) {
    while ( arg_len -- ) * arg_dst ++ = * arg_src ++;
    * arg_dst = (map_type) 0;
    return;
}
    
//-----------------------------------------------------------------------------
// function	map_xx_compare (internal)
//
// purpose	Compare the array keys in two AVL tree nodes.  This function
//		is not globally exported.  It is used only to establish the
//		comparison method for the AVL tree used to hold this mapping.
//
// arguments	1 (struct map_link *)
//		2 (struct map_link *)
//		3 (void *) extra pointer (not used)
//
// returns	(int) comparison status, -1, 0, 1
//-----------------------------------------------------------------------------
static
int
map_xx_compare (
    avl_link *		arg_one
    ,
    avl_link *		arg_two
    ,
    AVL *		arg_avl	// not used here, see map_xx_compare_alt
) {
    map_type *		ptr1	;
    map_type *		ptr2	;
    size_t		len1	;
    size_t		len2	;

    ptr1 = ( (struct map_link *) arg_one )->key;
    len1 = ( (struct map_link *) arg_one )->keylen;
    ptr2 = ( (struct map_link *) arg_two )->key;
    len2 = ( (struct map_link *) arg_two )->keylen;
    while ( len1 && len2 && * ptr1 == * ptr2 ) {
	-- len1;
	-- len2;
	++ ptr1;
	++ ptr2;
    }
    if ( len1 == 0 && len2 == 0 ) return 0;
    if ( len1 == 0 ) return -1;
    if ( len2 == 0 ) return 1;

#ifdef MAP_COMPARE_CAST
    return ( ((MAP_COMPARE_CAST)(* ptr1)) < ((MAP_COMPARE_CAST)(* ptr2)) ) ?
	-1 : ( ((MAP_COMPARE_CAST)(* ptr1)) > ((MAP_COMPARE_CAST)(* ptr2)) ) ? 1 : 0;
#else
    return ( * ptr1 < * ptr2 ) ? -1 : ( * ptr1 > * ptr2 ) ? 1 : 0;
#endif
}

//-----------------------------------------------------------------------------
// function	map_xx_compare_alt (internal)
//
// purpose	Compare the array keys in two AVL tree nodes.  This function
//		is not globally exported.  It is used only to establish the
//		comparison method for the AVL tree used to hold this mapping.
//
// arguments	1 (struct map_link *)
//		2 (struct map_link *)
//		3 (void *) alternate compare function
//
// returns	(int) comparison status, -1, 0, 1
//-----------------------------------------------------------------------------
static
int
map_xx_compare_alt (
    avl_link *		arg_one
    ,
    avl_link *		arg_two
    ,
    AVL *		arg_avl
) {
    if ( map_cmp_func( arg_avl ) ) {
	return ( * map_cmp_func( arg_avl ) ) (
	    ( (struct map_link *) arg_one )->key,
	    ( (struct map_link *) arg_one )->keylen,
	    ( (struct map_link *) arg_two )->key,
	    ( (struct map_link *) arg_two )->keylen
	);
    }
    return map_xx_compare( arg_one, arg_two, NULL );
}

//-----------------------------------------------------------------------------
// function	map_xx_new
//
// purpose	Create a new mapping with the specific array key type.
//
// arguments	1 (int *)() optional alternate comparison function
//
// returns	(MAP) handle for the new mapping.
//		(MAP) NULL = error creating new mapping
//-----------------------------------------------------------------------------
MAP
map_xx_new (
    int		(* arg_compare)(map_type *,map_type *)
)
{
    MAP		map_p		;

    //-- Allocate a structure for the root of the mapping.
    if ( ! ( map_p = malloc( sizeof (struct map_root) ) ) ) return NULL;

    //-- Record the key type.
    map_p->key_type = MAP_TYPE_CODE;

    //-- Let the AVL layer initialize the rest of it.
    if ( arg_compare ) {
	map_cmp_func( map_p ) = arg_compare;
	if ( ! avl_init( & (map_p->avl_tree), struct map_node, link, map_xx_compare_alt ) ) return NULL;
    } else {
	if ( ! avl_init( & (map_p->avl_tree), struct map_node, link, map_xx_compare ) ) return NULL;
    }

    //-- Return the handle to the new mapping.
    return map_p;
}

//-----------------------------------------------------------------------------
// function	map_xx_find
//
// purpose	Find a member with a matching key, or possibly a nearby key.
//
//		If the influential direction is none, return the first exact
//		match encountered, or return none.
//
//		If the influential direction is low, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest member in the
//		low direction.
//
//		If the influential direction is high, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest member in the
//		high direction.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) array key pointer
//		3 (size_t) array key length
//		4 (int) direction and near option, -2, -1, 0, 1, 2
//			-2 = choose lowest exact, or nearest lower
//			-1 = choose lowest exact only
//			 0 = choose first exact only
//			 1 = choose highest exact only
//			 2 = choose highest exact, or nearest higher
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no member found
//		(int)  0 = found a member with no data
//		(int) data type of found member
//-----------------------------------------------------------------------------
int
map_xx_find (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
    ,
    int			arg_dir_near
)
{
    struct map_node *	node_p		;
    struct map_link *	finder_p	;
    size_t		len		;

    //-- Make sure we have a correct mapping.
    if ( ! arg_map ) return -2;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return -2;

    //-- If the length is not given, determine it as zero terminated string.
    if ( arg_key_len == ~0U ) arg_key_len = map_xx_length( arg_key_ptr );

    //-- Set up a search link to pass the key.
    len = sizeof (struct map_link) + arg_key_len * sizeof (map_type);
    if ( ! ( finder_p = malloc( len ) ) ) return -1;
    finder_p->keylen = arg_key_len;
    map_xx_copy( finder_p->key, arg_key_ptr, arg_key_len );

    //-- Let the AVL layer do the search.
    node_p = avl_find( & (arg_map->avl_tree), finder_p, arg_dir_near );
    free( finder_p );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_xx_find_exact
//
// purpose	Find a member with an exactly matching key of the specific
//		array key type for this mapping.  If multiple members match,
//		this function does not define what particular member will be
//		selected.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no member found
//		(int)  0 = found a member with no data
//		(int) data type of found member
//-----------------------------------------------------------------------------
int
map_xx_find_exact (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    return map_xx_find( arg_map, arg_key_ptr, arg_key_len, 0 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_exact_lo
//
// purpose	Find a member with an exactly matching key of the specific
//		array key type for this mapping.  If there is more than
//		one member of the same key, select the one in the lowest
//		position.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no member found
//		(int)  0 = found a member with no data
//		(int) data type of found member
//-----------------------------------------------------------------------------
int
map_xx_find_exact_lo (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    return map_xx_find( arg_map, arg_key_ptr, arg_key_len, -1 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_lo_or_near
//
// purpose	Find a member with an exactly matching key of the specific
//		array key type for this mapping, or if there is no exactly
//		matching key, select the nearest lower key.  If there is
//		more than one exactly matching key, select the one in the
//		lowest position.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no member found
//		(int)  0 = found a member with no data
//		(int) data type of found member
//-----------------------------------------------------------------------------
int
map_xx_find_lo_or_near (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    return map_xx_find( arg_map, arg_key_ptr, arg_key_len, -2 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_exact_hi
//
// purpose	Find a member with an exactly matching key of the specific
//		array key type for this mapping.  If there is more than
//		one member of the same key, select the one in the highest
//		position.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no member found
//		(int)  0 = found a member with no data
//		(int) data type of found member
//-----------------------------------------------------------------------------
int
map_xx_find_exact_hi (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    return map_xx_find( arg_map, arg_key_ptr, arg_key_len, 1 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_hi_or_near
//
// purpose	Find a member with an exactly matching key of the specific
//		array key type for this mapping, or if there is no exactly
//		matching key, select the nearest higher key.  If there is
//		more than one exactly matching key, select the one in the
//		highest position.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no member found
//		(int)  0 = found a member with no data
//		(int) data type of found member
//-----------------------------------------------------------------------------
int
map_xx_find_hi_or_near (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    return map_xx_find( arg_map, arg_key_ptr, arg_key_len, 2 );
}

//-----------------------------------------------------------------------------
// function	map_xx_insert
//
// purpose	Insert a new member into a mapping with the specified key and
//		an initial undefined data value, unless one already exists
//		with that exact key, in which case it is made active.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int)  0 = found or inserted a member with no data
//		(int) >0 = data type of found member
//-----------------------------------------------------------------------------
int
map_xx_insert (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    struct map_node *	node_p	;
    int			rc	;

    //-- Make sure we have a correct mapping.
    if ( ! arg_map ) return -2;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return -2;

    //-- If the length is not given, determine it as zero terminated string.
    if ( arg_key_len == ~0U ) arg_key_len = map_xx_length( arg_key_ptr );

    //-- Allocate a combined node and key space.  The node already
    //-- has the space for one element of the key so it is not
    //-- necessary to add an extra one for the terminating element.
    if ( ! ( node_p = malloc( sizeof (struct map_node) + sizeof (map_type) * arg_key_len ) ) ) return -2;

    //-- Initialize the node.
    node_p->data.type = 0;
    node_p->link.keylen = arg_key_len;

    //-- Copy the key.
    map_xx_copy( node_p->link.key, arg_key_ptr, arg_key_len );

    //-- Insert node or return data type if already exists.
    rc = avl_insert( & (arg_map->avl_tree), node_p );
    if ( rc >= 0 ) return 0;
    free( node_p );
    return map_xx_find( arg_map, arg_key_ptr, arg_key_len, 0 );
}

//-----------------------------------------------------------------------------
// function	map_xx_insert_dup
//
// purpose	Insert a new member into a mapping with the specified key and
//		an initial undefined data value.  If a member already exists,
//		insert a duplicate.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) key array pointer
//		3 (size_t) key length
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = not inserted
//		(int)  1 = new unique member inserted
//		(int)  2 = duplicate member inserted
//-----------------------------------------------------------------------------
int
map_xx_insert_dup (
    MAP			arg_map
    ,
    const map_type *	arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    struct map_node *	node_p	;
    size_t		len	;
    int			rc	;

    //-- Make sure we have a correct mapping.
    if ( ! arg_map ) return -2;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return -2;

    //-- If the length is not given, determine it as zero terminated string.
    if ( arg_key_len == ~0U ) arg_key_len = map_xx_length( arg_key_ptr );

    //-- Allocate a combined node and key space.  The node already
    //-- has the space for one element of the key so it is not
    //-- necessary to add an extra one for the terminating element.
    len = ( sizeof (struct map_node) ) + ( arg_key_len * sizeof (map_type) );
    if ( ! ( node_p = malloc( len ) ) ) return -2;

    //-- Initialize the node.
    node_p->data.type = 0;
    node_p->link.keylen = arg_key_len;

    //-- Copy the key.
    map_xx_copy( node_p->link.key, arg_key_ptr, arg_key_len );

    //-- Insert the node.
    rc = avl_insert_dup( & (arg_map->avl_tree), node_p );

    //-- Return status based on insertion.
    if ( rc < 0 ) {
	free( node_p );
	return -1;
    }
    return rc;
}

//-----------------------------------------------------------------------------
// function	map_xx_key_ptr
//
// purpose	Get the key of the current member in the mapping, returning the
//		pointer to the copy used in the AVL tree.  The calling program
//		must not alter the contents of the key to avoid corrupting the
//		AVL tree.
//
// arguments	1 (MAP) mapping handle
//		2 (size_t *) where to store key array length
//
// returns	(const map_type *) pointer to shared key
//		(const map_type *) NULL = error, no map, bad type, no member
//-----------------------------------------------------------------------------
const map_type *
map_xx_key_ptr (
    MAP			arg_map
    ,
    size_t *		arg_key_len_p
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a correct mapping and current node.
    if ( ! arg_map ) return NULL;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return NULL;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return NULL;

    //-- If a key length storage location is given, store the length.
    if ( arg_key_len_p ) * arg_key_len_p = node_p->link.keylen;

    //-- Return the key pointer.
    return node_p->link.key;
}

//-----------------------------------------------------------------------------
// function	map_xx_key_dup
//
// purpose	Get the key of the current member in the mapping, returning an
//		allocated and duplicate copy.  The calling program must free
//		the allocated memory when use is concluded to avoid lossage
//		of memory.
//
// arguments	1 (MAP) mapping handle
//		2 (size_t *) where to store key array length
//
// returns	(map_type *) pointer to key duplicate copy
//		(map_type *) NULL = error, no map, bad type, no member
//-----------------------------------------------------------------------------
map_type *
map_xx_key_dup (
    MAP			arg_map
    ,
    size_t *		arg_key_len_p
)
{
    struct map_node *	node_p	;
    map_type *		key_p	;
    size_t		len	;

    //-- Make sure we have a correct mapping and current node.
    if ( ! arg_map ) return NULL;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return NULL;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return NULL;

    //-- Get key length.
    len = node_p->link.keylen;

    //-- Allocate space for key duplicate or return an error.
    if ( ! ( key_p = malloc( ( len + 1 ) * sizeof (map_type) ) ) ) return NULL;

    //-- If a key length storage location is given, store the length.
    if ( arg_key_len_p ) * arg_key_len_p = len;
    
    //-- Copy the key for duplication.
    map_xx_copy( key_p, node_p->link.key, len );

    //-- Return the pointer to the duplicate.
    return key_p;
}

//-----------------------------------------------------------------------------
// function	map_xx_key_copy
//
// purpose	Get the key of the current member in the mapping, copying it to
//		the specified destination area.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type *) where to store key
//		3 (size_t) maximum length of storage area
//
// returns	(size_t) length of key being returned (even if not all copied)
//		(size_t) ~0U = error, no map, bad type, no member
//-----------------------------------------------------------------------------
size_t
map_xx_key_copy (
    MAP			arg_map
    ,
    map_type *		arg_key_ptr
    ,
    size_t		arg_key_len
)
{
    struct map_node *	node_p	;
    size_t		len	;

    //-- Make sure we have a correct mapping and current node.
    if ( ! arg_map ) return ~0U;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return ~0U;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return ~0U;

    //-- Copy the key to the destination, truncated if necessary.
    len = node_p->link.keylen;
    if ( len > arg_key_len ) len = arg_key_len;
    map_xx_copy( arg_key_ptr, node_p->link.key, len );

    //-- Return length for convenience.
    return node_p->link.keylen;
}

//-----------------------------------------------------------------------------
// function	map_xx_key_len
//
// purpose	Get the length of a key value of the current member in the
//		mapping.
//
// arguments	1 (MAP) mapping handle
//
// returns	(size_t) key length
//		(size_t) ~0U = error, no map, bad type, no member
//-----------------------------------------------------------------------------
size_t
map_xx_key_len (
    MAP			arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a correct mapping and current node.
    if ( ! arg_map ) return ~0U;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return ~0U;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return ~0U;

    //-- Return the key length.
    return node_p->link.keylen;
}

//-----------------------------------------------------------------------------
// function	map_xx_store
//
// purpose	Store the given data into the current member in the mapping.
//		String/array data is stored as a duplicate so the original
//		data can be modified or the space reused or freed after
//		this function returns.
//
// arguments	1 (MAP) mapping handle
//		2 (const map_type *) data value to store
//		3 (size_t) length of data value
//
// returns	(int) -2 = error, no map, no member, out of memory
//		(int)  0 = success, data stored
//-----------------------------------------------------------------------------
int
map_xx_store (
    MAP			arg_map
    ,
    const map_type *	arg_data_ptr
    ,
    size_t		arg_data_len
)
{
    struct map_node *	node_p	;
    map_type *	data_p	;
    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return -2;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return -2;

    //-- If the length is not given, determine it as zero terminated string.
    if ( arg_data_len == ~0U ) arg_data_len = map_xx_length( arg_data_ptr );

    //-- Allocate space for the data to be duplicated.
    if ( ! ( data_p = malloc( ( arg_data_len + 1 ) * sizeof (map_type) ) ) ) return -2;

    //-- Copy the data to the duplication space.
    map_xx_copy( data_p, arg_data_ptr, arg_data_len );

    //-- Store data in the node.
    node_p->data.mix.string.ptr = (void *) data_p;
    node_p->data.mix.string.len = arg_data_len;
    node_p->data.mix.string.dup = 1;
    node_p->data.type = MAP_TYPE_CODE;

    //-- Return a success status.
    return 0;
}

//-----------------------------------------------------------------------------
// function	map_xx_store_share
//
// purpose	Store the given data into the current member in the mapping.
//		The data pointer is stored as given and a duplicate is not
//		made.  This results in data being shared between the mapping
//		and the calling program.  The calling program may modify the
//		data in place, and is responsible for the consequences of
//		doing so since it changes the data (but not the type) stored
//		in the mapping.  Such a change would also affect references
//		of fetched data pointers.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type *) data value to store
//		3 (size_t) length of data value
//
// returns	(int) -2 = error, no map, no member
//		(int)  0 = success, data stored
//-----------------------------------------------------------------------------
int
map_xx_store_share (
    MAP			arg_map
    ,
    map_type *		arg_data_ptr
    ,
    size_t		arg_data_len
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return -2;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return -2;

    //-- Store the data type, length, and pointer.
    node_p->data.mix.string.ptr = (void *) arg_data_ptr;
    node_p->data.mix.string.len = arg_data_len;
    node_p->data.mix.string.dup = 0;
    node_p->data.type = MAP_TYPE_CODE;

    //-- Return a success status.
    return 0;
}

//-----------------------------------------------------------------------------
// function	map_xx_fetch_ptr
//
// purpose	Fetch the data from the current member in the mapping.
//
// arguments	1 (MAP) mapping handle
//		2 (size_t *) pointer where to store data length
//
// returns	(const map_type *) pointer to data string array
//		(const map_type *) NULL = error, no map, no member, bad type
//-----------------------------------------------------------------------------
const map_type *
map_xx_fetch_ptr (
    MAP			arg_map
    ,
    size_t *		arg_data_len_p
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return NULL;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return NULL;

    //-- Make sure data type is correct.
    if ( node_p->data.type != MAP_TYPE_CODE ) return NULL;

    //-- If a data length storage location is given, store the length.
    if ( arg_data_len_p ) * arg_data_len_p = node_p->data.mix.string.len;

    //-- Return the data pointer.
    return (const map_type *) ( node_p->data.mix.string.ptr );
}

//-----------------------------------------------------------------------------
// function	map_xx_fetch_dup
//
// purpose	Fetch the data from the current member in the mapping.
//
// arguments	1 (MAP) mapping handle
//		2 (size_t *) pointer where to store data length
//
// returns	(map_type *) pointer to duplicate data string array
//		(map_type *) NULL = error, no map, no member, bad type
//-----------------------------------------------------------------------------
map_type *
map_xx_fetch_dup (
    MAP			arg_map
    ,
    size_t *		arg_data_len_p
)
{
    struct map_node *	node_p	;
    map_type *		data_p	;
    size_t		len	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return NULL;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return NULL;

    //-- Make sure data type is correct.
    if ( node_p->data.type != MAP_TYPE_CODE ) return NULL;

    //-- Get data length.
    len = node_p->data.mix.string.len;

    //-- Allocate space for data duplicate or return an error.
    if ( ! ( data_p = malloc( ( len + 1 ) * sizeof (map_type) ) ) ) return NULL;

    //-- If a data length storage location is given, store the length.
    if ( arg_data_len_p ) * arg_data_len_p = len;

    //-- Copy the data for duplication.
    map_xx_copy( data_p, (map_type *) ( node_p->data.mix.string.ptr ), len );

    //-- Return the pointer to the duplicate.
    return data_p;
}

//-----------------------------------------------------------------------------
// function	map_xx_fetch_copy
//
// purpose	Fetch the data from the current member in the mapping.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type *) pointer where to store data
//		3 (size_t) maximum length to store
//
// returns	(size_t) ~0U = error, no map, no member, bad type
//		(size_t) length of full string (copy may be truncated)
//-----------------------------------------------------------------------------
size_t
map_xx_fetch_copy (
    MAP			arg_map
    ,
    map_type *		arg_data_ptr
    ,
    size_t		arg_data_len
)
{
    struct map_node *	node_p	;
    size_t		len	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return ~0U;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return ~0U;

    //-- Make sure data type is correct.
    if ( node_p->data.type != MAP_TYPE_CODE ) return ~0U;

    //-- Copy the data to the destination, truncated if necessary.
    len = node_p->data.mix.string.len;
    if ( len > arg_data_len ) len = arg_data_len;
    map_xx_copy( arg_data_ptr, (map_type *) ( node_p->data.mix.string.ptr ), len );

    //-- Return original length for convenience.
    return node_p->data.mix.string.len;
}

//-----------------------------------------------------------------------------
// function	map_xx_fetch_len
//
// purpose	Fetch the data from the current member in the mapping.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type * *) pointer where to store data
//
// returns	(size_t) ~0U = error, no map, no member, bad type
//		(size_t) length of full string (copy may be truncated)
//-----------------------------------------------------------------------------
size_t
map_xx_fetch_len (
    MAP			arg_map
    ,
    map_type * *	arg_data_p
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return ~0U;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return ~0U;

    //-- Make sure data type is correct.
    if ( node_p->data.type != MAP_TYPE_CODE ) return ~0U;

    //-- Return length.
    return node_p->data.mix.string.len;
}
