//-----------------------------------------------------------------------------
// Copyright  2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// program	wordstat
//
// purpose	Count the frequency of different word items in the input.
//
// syntax	wordstat [-q] ["sepchars"] < infile
//-----------------------------------------------------------------------------
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <map.h>

#define WORDSIZE	256

//-----------------------------------------------------------------------------
// function	readword
//-----------------------------------------------------------------------------
static inline
int
readword (
    char *	arg_buffer
,
    size_t	arg_length
,
    char *	arg_seps
)
{
    char *	p	;
    char *	q	;
    int		c	;

    //-- Note where the end of the buffer is.
    q = ( p = arg_buffer ) + arg_length - 1;

    //-- Skip all leading separator characters.
    while ( (c = getc( stdin )) != EOF && arg_seps[ c ] ) {
    }

    //-- Get characters until end of word, store until full.
    while ( c != EOF && arg_seps[ c ] == 0 ) {
	if ( p < q ) * p ++ = c;
	c = getc( stdin );
    }
    * p = 0;

    //-- Return 0 if EOF, 1 if a word found.
    return ( c == EOF );
}

//-----------------------------------------------------------------------------
// function	main
//-----------------------------------------------------------------------------
int
main (
    int		argc
,
    char * *	argv
)
{
    MAP			by_word		;
    MAP			by_count	;

    unsigned long	wcount		;

    char		word		[ WORDSIZE ];
    char		seps		[ 256 ];


    //------------------------------------------------------------------------
    // Initialize the separator table that is given to readword().  This table
    // tells it what characters separate each word.  Control characters are
    // always placed in the table.  If no explicit characters are given in the
    // third argument then all space and punctuation characters are added by
    // default.  Otherwise just the given characters are added.
    //------------------------------------------------------------------------
    {
	int i;

	//-- Initialize the table with just control characters.
	for ( i = 0; i < 256; ++ i ) seps[ i ] = iscntrl( i );

	if ( argc > 1 && argv[ 1 ] ) {
	    char * p;

	    //-- Add the characters from the given arguments.
	    p = argv[ 1 ];
	    while ( * p ) seps[ 255 & * p ++ ] = 1;

	} else {

	    //-- Add space and punctuation characters.
	    for ( i = 0; i < 256; ++ i ) {
		if ( ispunct( i ) || isspace( i ) ) seps[ i ] = 1;
	    }
	}
    }

    //---------------------
    // Create the mappings.
    //---------------------
    by_word = map_str_new( NULL );
    by_count = map_ul_new( NULL );

    //----------------------------------------------------------
    // Read each word and insert or increment it in the mapping.
    //----------------------------------------------------------
    wcount = 0;
    while ( 0 == readword( word, WORDSIZE, seps ) ) {
	unsigned long count;
	if ( map_str_find_exact( by_word, word ) > 0 ) {
	    count = map_ul_fetch( by_word );
	} else {
	    map_str_insert( by_word, word );
	    count = 0;
	}
	++ count;
	map_ul_store( by_word, count );
	++ wcount;
    }

    //-------------------------------------------------------
    // Transfer all the words from word order to count order.
    //-------------------------------------------------------
    if ( map_first( by_word ) > 0 ) do {
	map_ul_insert_dup( by_count, map_ul_fetch( by_word ) );
	map_str_store( by_count, map_str_key_ptr( by_word, NULL ) );
    } while ( map_forward( by_word ) > 0 );
    map_destroy( by_word );
    
    //-----------------------------------------------
    // Print all the words and counts in count order.
    //-----------------------------------------------
    if ( map_first( by_count ) > 0 ) do {
	printf( "%11lu %s\n",
		map_ul_key( by_count ),
		map_str_fetch_ptr( by_count, NULL )
	);
    } while ( map_forward( by_count ) > 0 );
    map_destroy( by_count );

    printf( "total words read = %lu\n",wcount);
    return 0;
}
